/*
    Copyright (C) 2007 Red Hat Software

    This file is part of Red Hat Messaging.
    
    Red Hat Messaging is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
    USA

    The GNU Lesser General Public License is available in the file COPYING.
*/

#include "BdbMessageStore.h"
#include <qpid/broker/RecoveryManager.h>
#include <qpid/broker/BrokerMessage.h>
#include <qpid/broker/BrokerMessageMessage.h>
#include <qpid/framing/Buffer.h>
#include <sstream>

using namespace qpid::bdbstore;
using namespace qpid::broker;
using boost::static_pointer_cast;

using std::auto_ptr;
using qpid::framing::Buffer;
using qpid::framing::FieldTable;
using qpid::sys::Mutex;

static const u_int8_t MESSAGE_MESSAGE = 1;
static const u_int8_t BASIC_MESSAGE = 2;

BdbMessageStore::BdbMessageStore(const char* envpath) : env(0), 
                                                        queueDb(&env, 0), 
                                                        messageDb(&env, 0), 
                                                        mappingDb(&env, 0), 
                                                        enqueueXidDb(&env, 0), 
                                                        dequeueXidDb(&env, 0), 
                                                        prepareXidDb(&env, 0), 
                                                        nextQueueId(1), 
                                                        nextMessageId(1),
                                                        enqueueHelper(mappingDb, "Enqueue failed"),
                                                        enqueueXidHelper(enqueueXidDb, "enqueue", "Marking of 2pc enqueue failed"),
                                                        dequeueXidHelper(dequeueXidDb, "dequeue", "Marking of 2pc dequeue failed")
 
{
    try {
        env.open(envpath, DB_THREAD | DB_CREATE | DB_INIT_TXN | DB_INIT_LOCK | DB_INIT_LOG | DB_INIT_MPOOL | DB_USE_ENVIRON, 0);
        
        u_int32_t flags = DB_CREATE | DB_THREAD;

        TxnCtxt txn;
        txn.begin(env);
        queueDb.open(txn.get(), "queues.db", 0, DB_BTREE, flags, 0);
        messageDb.open(txn.get(), "messages.db", 0, DB_BTREE, flags, 0);
        
        mappingDb.set_flags(DB_DUPSORT);
        mappingDb.open(txn.get(), "mappings.db", 0, DB_BTREE, flags, 0);

        enqueueXidDb.set_flags(DB_DUPSORT);
        enqueueXidDb.open(txn.get(), "enqueue_xid.db", 0, DB_BTREE, flags, 0);

        dequeueXidDb.set_flags(DB_DUPSORT);
        dequeueXidDb.open(txn.get(), "dequeue_xid.db", 0, DB_BTREE, flags, 0);

        prepareXidDb.open(txn.get(), "prepare_xid.db", 0, DB_BTREE, flags, 0);

        txn.commit();
    } catch (DbException& e) {
        THROW_STORE_EXCEPTION_2("Error opening databases", e);
    } catch (std::exception& e) {
        // std::cout << "Error opening databases: " << e.what() << std::endl;
        throw e;
    }
}

BdbMessageStore::~BdbMessageStore() 
{
    try {
        queueDb.close(0);
        messageDb.close(0);
        mappingDb.close(0);
        enqueueXidDb.close(0);
        dequeueXidDb.close(0);
        prepareXidDb.close(0);
    } catch (DbException& e) {
        THROW_STORE_EXCEPTION_2("Error closing databases", e);
    } catch (std::exception& e) {
        // std::cout << "Error closing databases: " << e.what() << std::endl;
        throw e;
    }   
}

void BdbMessageStore::truncate()
{
    DbTxn* txn;
    env.txn_begin(0, &txn, 0);
    u_int32_t count;

    queueDb.truncate(txn, &count, 0);
    messageDb.truncate(txn, &count, 0);
    mappingDb.truncate(txn, &count, 0);
    enqueueXidDb.truncate(txn, &count, 0);
    dequeueXidDb.truncate(txn, &count, 0);
    prepareXidDb.truncate(txn, &count, 0);
    txn->commit(0);        
}

void BdbMessageStore::create(const PersistableQueue& queue)
{
    if (queue.getPersistenceId()) {
        THROW_STORE_EXCEPTION("Queue already created: " + queue.getName());
    }
    u_int64_t queueId (getNextQueueId());
    Dbt key(&queueId, sizeof(queueId));

    //value includes the internal id and the settings    
    Buffer buffer(queue.encodedSize() + sizeof(u_int64_t));
    queue.encode(buffer);
    buffer.flip();
    Dbt value (buffer.start(), buffer.available());
    
    try {
        int status = queueDb.put(0, &key, &value, DB_NOOVERWRITE | DB_AUTO_COMMIT);
        queue.setPersistenceId(queueId);
        if (status == DB_KEYEXIST) {
            THROW_STORE_EXCEPTION("Queue already exists: " + queue.getName());
        }
    } catch (DbException& e) {
        THROW_STORE_EXCEPTION_2("Error creating queue named " + queue.getName(), e);
    }
}

void BdbMessageStore::destroy(const PersistableQueue& queue)
{
    u_int64_t id = queue.getPersistenceId();
    Dbt key(&id, sizeof(id));

    queueDb.del(0, &key, DB_AUTO_COMMIT);
}

void BdbMessageStore::create(const PersistableExchange&)
{
    //TODO
}
void BdbMessageStore::destroy(const PersistableExchange&)
{
    //TODO
}

void BdbMessageStore::recover(RecoveryManager& registry)
{
    recoverXids();

    queue_index index;//id->queue
    //read all queues
    recoverQueues(registry, index);
    //read all messages
    recoverMessages(registry, index);
}

void BdbMessageStore::recoverQueues(RecoveryManager& registry, queue_index& index)
{
    CursorWrapper queues;
    queues.open(queueDb, 0);

    u_int64_t maxQueueId(1);

    uint64_t id(0);
    Dbt key(&id, sizeof(id));
    key.set_ulen(sizeof(u_int64_t));
    key.set_flags(DB_DBT_USERMEM);

    Dbt value;
    //read all queues
    for (int status = queues->get(&key, &value, DB_NEXT); status == 0; status = queues->get(&key, &value, DB_NEXT)) {
        Buffer buffer(reinterpret_cast<char*>(value.get_data()), value.get_size());
        //create a Queue instance
        RecoverableQueue::shared_ptr queue = registry.recoverQueue(buffer);
        //set the persistenceId and update max as required
        queue->setPersistenceId(id);
        index[id] = queue;
        if (id > maxQueueId) {
            maxQueueId = id;
        }
    }
    nextQueueId = maxQueueId + 1;
}

void BdbMessageStore::recoverMessages(RecoveryManager& recovery, queue_index& index)
{
    CursorWrapper messages;
    messages.open(messageDb, 0);

    Dbt key;
    //convert key to u_int64_t (i.e. id)
    u_int64_t id;
    key.set_data(&id);
    key.set_ulen(sizeof(u_int64_t));
    key.set_flags(DB_DBT_USERMEM);

    size_t preamble_length = sizeof(u_int32_t)/*header size*/ + sizeof(u_int8_t)/*type code*/;

    BufferValue value(preamble_length, 0);

    u_int64_t maxMessageId(1);

    for (int status = messages->get(&key, &value, DB_NEXT); status == 0; status = messages->get(&key, &value, DB_NEXT)) {
        //read header only to begin with
        value.buffer.getOctet();//TODO, remove - no longer used
        u_int32_t headerSize = value.buffer.getLong();
        value.buffer.clear();

        BufferValue header(headerSize, preamble_length);
        messages->get(&key, &header, DB_CURRENT);

        RecoverableMessage::shared_ptr msg = recovery.recoverMessage(header.buffer);
        msg->setPersistenceId(id);
        
        u_int32_t contentOffset = headerSize + preamble_length;
        u_int64_t contentSize = getRecordSize(messageDb, key) - contentOffset;
        if (msg->loadContent(contentSize)) {
            //now read the content
            BufferValue content(contentSize, contentOffset);
            messages->get(&key, &content, DB_CURRENT);
            msg->decodeContent(content.buffer);
        }
        
        //find all the queues into which this message has been enqueued
        if (enqueueMessage(key, msg, index) == 0) {
            //message not referenced anywhere - can delete
            messages->del(0);
        } else {
            if (id > maxMessageId) {
                maxMessageId = id;
            }
        }
    }
    nextMessageId = maxMessageId + 1;
}

int BdbMessageStore::enqueueMessage(Dbt& msgId, RecoverableMessage::shared_ptr& msg, queue_index& index)
{
    CursorWrapper mappings;
    mappings.open(mappingDb, 0);

    Dbt value;
    u_int64_t queueId;
    value.set_data(&queueId);
    value.set_ulen(sizeof(u_int64_t));
    value.set_flags(DB_DBT_USERMEM);

    int count(0);
    for (int status = mappings->get(&msgId, &value, DB_SET); status == 0; status = mappings->get(&msgId, &value, DB_NEXT_DUP)) {
        RecoverableQueue::shared_ptr queue = index[queueId];
        if (!queue) {
            THROW_STORE_EXCEPTION("Missing queue");
        }
        queue->recover(msg);
        count++;
    }
    return count;
}

void BdbMessageStore::recoverXids()
{
    std::set<string> prepared;
    std::set<string> known;

    readXids(prepareXidDb, prepared);
    readXids(enqueueXidDb, known);
    readXids(dequeueXidDb, known);

    //abort all known but unprepared xids:
    for (std::set<string>::iterator i = known.begin(); i != known.end(); i++) {
        if (prepared.find(*i) == prepared.end()) {
            TPCTxnCtxt txn(*i);
            completed(txn, dequeueXidDb, enqueueXidDb);
        }
    }
}

void BdbMessageStore::readXids(Db& db, std::set<string>& xids)
{
    CursorWrapper c;
    c.open(db, 0);

    Dbt key;
    Dbt value;
    for (int status = c->get(&key, &value, DB_NEXT_NODUP); status == 0; status = c->get(&key, &value, DB_NEXT_NODUP)) {
        std::string xid;
        xid.assign(reinterpret_cast<char*>(key.get_data()), key.get_size());
        xids.insert(xid);
    }
}

void BdbMessageStore::collectPreparedXids(std::set<string>& xids)
{
    readXids(prepareXidDb, xids);    
}

void BdbMessageStore::stage(PersistableMessage& msg)
{
    TxnCtxt txn;
    txn.begin(env);

    u_int64_t messageId (msg.getPersistenceId());
    if (messageId == 0) {
        try {
            Dbt key (&messageId, sizeof(messageId));
            messageId = getNextMessageId();
            store(txn.get(), key, msg);
            msg.setPersistenceId(messageId);
            txn.commit();
        } catch (std::exception& e) {
            txn.abort();
            throw e;
        }
    }        
}
void BdbMessageStore::destroy(PersistableMessage& msg)
{
    u_int64_t messageId (msg.getPersistenceId());
    if (messageId) {
        Dbt key (&messageId, sizeof(messageId));
        TxnCtxt txn;
        txn.begin(env);
        try {
            deleteIfUnused(txn.get(), key);
            txn.commit();
        } catch (DbException& e) {
            txn.abort();
            THROW_STORE_EXCEPTION_2("Error destroying message", e);
        }            
    }
}

u_int64_t BdbMessageStore::getRecordSize(Db& db, Dbt& key)
{
    Dbt peek;
    peek.set_flags(DB_DBT_USERMEM);
    peek.set_ulen(0);
    try {
        int status = db.get(0, &key, &peek, 0);
        if (status != DB_BUFFER_SMALL) {
            THROW_STORE_EXCEPTION(appendBdbError("Unexpected status code when determining record length: ", status));
        }
    } catch (DbMemoryException& expected) {
        //api doc indicates may throw exception instead of status = DB_BUFFER_SMALL;
    }
    return peek.get_size();
}

void BdbMessageStore::appendContent(PersistableMessage& msg, const std::string& data)
{
    //TODO
    u_int64_t messageId (msg.getPersistenceId());
    if (messageId != 0) {
        try {
            Dbt key (&messageId, sizeof(messageId));
            u_int64_t offset = getRecordSize(messageDb, key);
            const int size(data.length());

            //don't want to have to copy this and shouldn't need to as
            //it will be used for reading only. but have to cast away
            //the const-ness, which is nasty...
            //alternative involves copying:
            //char buffer[size];
            //data.copy(buffer, size);
            Dbt value((void*) data.data(), size);
            value.set_flags(DB_DBT_USERMEM | DB_DBT_PARTIAL);
            value.set_doff(offset);
            value.set_dlen(size);
            messageDb.put(0, &key, &value, DB_AUTO_COMMIT);
        } catch (DbException& e) {
            THROW_STORE_EXCEPTION_2("Error appending content", e);
        }
    } else {
        THROW_STORE_EXCEPTION("Cannot append content. Message not known to store!");
    }    
}
void BdbMessageStore::loadContent(PersistableMessage& msg, std::string& data, u_int64_t offset, u_int32_t length)
{
    u_int64_t realOffset = offset + sizeof(u_int32_t)/*header length*/ + sizeof(u_int8_t)/*type code*/+ msg.encodedHeaderSize();
    u_int64_t messageId (msg.getPersistenceId());
    if (messageId != 0) {
        try {
            Dbt key (&messageId, sizeof(messageId));
            char *buffer = new char[length];
            Dbt value(buffer, length);
            value.set_flags(DB_DBT_USERMEM | DB_DBT_PARTIAL);
            value.set_ulen(length);
            value.set_doff(realOffset);
            value.set_dlen(length);
            int status = messageDb.get(0, &key, &value, 0); 
            if (status == DB_NOTFOUND) {
		delete [] buffer;
                THROW_STORE_EXCEPTION("Cannot load content. Message not known to store!");                
            } else {
                data.assign(buffer, value.get_size());
		delete [] buffer;
            }
        } catch (DbException& e) {
            THROW_STORE_EXCEPTION_2("Error loading content", e);
        }
    } else {
        THROW_STORE_EXCEPTION("Cannot load content. Message not known to store!");
    }
}

void BdbMessageStore::enqueue(TransactionContext* ctxt, PersistableMessage& msg, const PersistableQueue& queue)
{
    u_int64_t queueId (queue.getPersistenceId());
    u_int64_t messageId (msg.getPersistenceId());
    if (queueId == 0) {
        THROW_STORE_EXCEPTION("Queue not created: " + queue.getName());
    }
    Dbt key (&messageId, sizeof(messageId));
    Dbt value (&queueId, sizeof(queueId));

    TxnCtxt implicit;
    TxnCtxt* txn = 0;
    if (ctxt) {
        txn = check(ctxt);
    } else {
        txn = &implicit;
        txn->begin(env);
    }

    try {
        if (messageId == 0) {
            messageId = getNextMessageId();
            store(txn->get(), key, msg);
            msg.setPersistenceId(messageId);
        }
        enqueueHelper.put(txn->get(), key, value);
        if (txn->isTPC()) {
            enqueueXidHelper.put(dynamic_cast<TPCTxnCtxt&>(*txn), messageId, queueId);
        }
        
        if (!ctxt) txn->commit();
    } catch (std::exception& e) {
        if (!ctxt) txn->abort();
        throw e;
    }
}

void BdbMessageStore::store(DbTxn* txn, Dbt& messageId, PersistableMessage& message)
{
    Buffer buffer(message.encodedSize() + sizeof(u_int32_t)/*header length*/ + sizeof(u_int8_t)/*type code*/);
    //what type of message is it:
    if (dynamic_cast<MessageMessage*>(&message)) {
        buffer.putOctet(MESSAGE_MESSAGE);
    } else {
        buffer.putOctet(BASIC_MESSAGE);
    }
    u_int32_t headerSize = message.encodedHeaderSize();
    buffer.putLong(headerSize);
    message.encode(buffer);
    buffer.flip();

    Dbt data(buffer.start(), buffer.available());
    try {
        messageDb.put(txn, &messageId, &data, DB_NOOVERWRITE);
    } catch (DbException& e) {
        THROW_STORE_EXCEPTION_2("Error storing message", e);
    }
}

void BdbMessageStore::dequeue(TransactionContext* ctxt, PersistableMessage& msg, const PersistableQueue& queue)
{    
    u_int64_t queueId (queue.getPersistenceId());
    u_int64_t messageId (msg.getPersistenceId());
    if (messageId == 0) {
        THROW_STORE_EXCEPTION("Error dequeing message, persistence id not set");
    }
    if (queueId == 0) {
        THROW_STORE_EXCEPTION("Queue not created: " + queue.getName());
    }

    TxnCtxt implicit;
    TxnCtxt* txn = 0;
    if (ctxt) {
        txn = check(ctxt);
    } else {
        txn = &implicit;
        txn->begin(env);
    }
    
    try {
        if (txn->isTPC()) {
            //if this is part of a 2pc transaction, then only record the dequeue now,
            //it will be applied on commit
            dequeueXidHelper.put(dynamic_cast<TPCTxnCtxt&>(*txn), messageId, queueId);
        } else {
            Dbt key (&messageId, sizeof(messageId));
            Dbt value (&queueId, sizeof(queueId));
            
            if (dequeue(txn->get(), key, value)) {
                msg.setPersistenceId(0);//clear id as we have now removed the message from the store
            }
        }
        if (!ctxt) txn->commit();
        
    } catch (DbException& e) {
        if (!ctxt) txn->abort();
        THROW_STORE_EXCEPTION_2("Error dequeing message", e);
    } catch (std::exception& e) {
        if (!ctxt) txn->abort();
        throw e;
    }   
}

bool BdbMessageStore::dequeue(DbTxn* txn, Dbt& messageId, Dbt& queueId)
{
    CursorWrapper cursor;
    cursor.open(mappingDb, txn);

    int status = cursor->get(&messageId, &queueId, DB_GET_BOTH | DB_RMW);
    if (status == 0) {
        cursor->del(0);
    } else if (status == DB_NOTFOUND ) {
        THROW_STORE_EXCEPTION("Can't find record mapping message to queue");
    } else {
        THROW_STORE_EXCEPTION("Dequeue failed with status = " + status);
    }

    return deleteIfUnused(cursor, txn, messageId);
}

bool BdbMessageStore::deleteIfUnused(DbTxn* txn, Dbt& messageId)
{
    CursorWrapper cursor;
    cursor.open(mappingDb, txn);

    return deleteIfUnused(cursor, txn, messageId);
}

bool BdbMessageStore::deleteIfUnused(CursorWrapper& cursor, DbTxn* txn, Dbt& messageId)
{
    Dbt empty;
    int status = cursor->get(&messageId, &empty, DB_SET);
    if (status == DB_NOTFOUND) {
        messageDb.del(txn, &messageId, 0);
        return true;
    } else if (status == 0) {
        return false;
    } else {
        THROW_STORE_EXCEPTION("Dequeue failed with status = " + status);
    }
}

void BdbMessageStore::completed(TPCTxnCtxt& txn, Db& discard, Db& apply)
{
    if (!txn.get()) txn.begin(env);

    string xid(txn.getXid());
    try {
        Dbt key ((void*) xid.data(), xid.length());

        //scroll through all records matching xid in apply and dequeue
        //using the message and queue id encoded in each value
        CursorWrapper c;
        c.open(apply, txn.get());

        u_int64_t ids[2];
        ids[0] = ids[1] = 0;
        Dbt value(&ids, sizeof(ids));
        value.set_ulen(sizeof(ids));
        value.set_flags(DB_DBT_USERMEM);

        Dbt messageId(&ids[0], sizeof(ids[0]));
        Dbt queueId(&ids[1], sizeof(ids[1]));

        for (int status = c->get(&key, &value, DB_SET); status == 0; status = c->get(&key, &value, DB_NEXT_DUP)) {
            dequeue(txn.get(), messageId, queueId);
        }
        c.close();

        //delete all records matching xid
        prepareXidDb.del(txn.get(), &key, 0);
        discard.del(txn.get(), &key, 0);
        apply.del(txn.get(), &key, 0);

        txn.commit();
    } catch (std::exception& e) {
        std::cout << "Error completing xid " << xid << ": " << e.what() << std::endl;
        txn.abort();
        throw e;
    }
}

auto_ptr<TransactionContext> BdbMessageStore::begin() 
{
    TxnCtxt* txn(new TxnCtxt());
    txn->begin(env);
    return auto_ptr<TransactionContext>(txn);
}

std::auto_ptr<qpid::broker::TPCTransactionContext> BdbMessageStore::begin(const std::string& xid)
{
    TPCTxnCtxt* txn(new TPCTxnCtxt(xid));
    txn->begin(env);
    return auto_ptr<TPCTransactionContext>(txn);
}

void BdbMessageStore::prepare(qpid::broker::TPCTransactionContext& ctxt)
{
    TPCTxnCtxt* txn = dynamic_cast<TPCTxnCtxt*>(&ctxt);
    if(!txn) throw InvalidTransactionContextException();
    
    try {
        u_int8_t dummy(1);
        string xid(txn->getXid());
        Dbt key ((void*) xid.data(), xid.length());
        Dbt value(&dummy, sizeof(dummy));

        prepareXidDb.put(txn->get(), &key, &value, 0);

        txn->commit();
    } catch (std::exception& e) {
        txn->abort();
        throw e;
    }

}

void BdbMessageStore::commit(TransactionContext& ctxt) 
{
    TxnCtxt* txn(check(&ctxt));
    if (txn->isTPC()) {
        completed(*dynamic_cast<TPCTxnCtxt*>(txn), enqueueXidDb, dequeueXidDb);        
    } else {
        txn->commit();
    }
}

void BdbMessageStore::abort(TransactionContext& ctxt) 
{
    TxnCtxt* txn(check(&ctxt));
    if (txn->isTPC()) {
        completed(*dynamic_cast<TPCTxnCtxt*>(txn), dequeueXidDb, enqueueXidDb);
    } else {
        txn->abort();
    }
}

TxnCtxt* BdbMessageStore::check(TransactionContext* ctxt)
{
    TxnCtxt* txn = dynamic_cast<TxnCtxt*>(ctxt);
    if(!txn) throw InvalidTransactionContextException();
    return txn;
}

u_int64_t BdbMessageStore::getNextQueueId()
{
    Mutex::ScopedLock guard(sequenceLock);
    return nextQueueId++;
}

u_int64_t BdbMessageStore::getNextMessageId()
{
    Mutex::ScopedLock guard(sequenceLock);
    return nextMessageId++;
}

BufferValue::BufferValue(u_int32_t size, u_int64_t offset) : buffer(size) {
    set_data(buffer.start());
    set_size(buffer.available());
    set_flags(DB_DBT_USERMEM | DB_DBT_PARTIAL);
    set_doff(offset);
    set_dlen(buffer.available());
    set_ulen(buffer.available());        
}

extern "C" MessageStore* create()
{
    // std::cout << "creating BdbMessageStore" <<std::endl;
    return new BdbMessageStore();
}

extern "C" void destroy(MessageStore* store)
{
    // std::cout << "destroying BdbMessageStore" <<std::endl;
    delete store;
}

string qpid::bdbstore::appendLocationInfo(const string& text, const string& file, int line)
{
    if (!file.empty() && line) {
        std::stringstream message;
        message << text << " [" << file << ", line " << line << "]";
        return message.str();
    } else {
        return text;
    }
}

string qpid::bdbstore::appendBdbError(const string& text, int code)
{
    string msg(text);
    msg += DbEnv::strerror(code);
    return msg;
}

XidOpRecorder::XidOpRecorder(Db& _db, const string& _name, const string& _error) : DupKeyRecorder(_db,_error), name(_name) {}

void XidOpRecorder::put(TPCTxnCtxt& txn, u_int64_t messageId, u_int64_t queueId)
{
    u_int64_t ids[] = {messageId, queueId};
    string xid(txn.getXid());//take a copy to avoid casting away 'const'ness, api doesn't recognise that key won't be altered
    Dbt key((void*) xid.data(), xid.length());
    Dbt value(&ids, sizeof(ids));
    DupKeyRecorder::put(txn.get(), key, value);
}

DupKeyRecorder::DupKeyRecorder(Db& _db, const std::string& _error) : db(_db), error(_error) {}

void DupKeyRecorder::put(DbTxn* txn, Dbt& key, Dbt& value)
{
    try {
        int status = db.put(txn, &key, &value, DB_NODUPDATA);
        if (status == DB_KEYEXIST) {
            THROW_STORE_EXCEPTION(error + ": duplicate data");
        } else if (status) {
            THROW_STORE_EXCEPTION(appendBdbError(error + ": status = ", status));
        }
    } catch (DbException& e) {
        THROW_STORE_EXCEPTION_2(error, e);
    }
}
