/*
    Copyright (C) 2007 Red Hat Software

    This file is part of Red Hat Messaging.
    
    Red Hat Messaging is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
    USA

    The GNU Lesser General Public License is available in the file COPYING.
*/

#ifndef _BdbMessageStore_
#define _BdbMessageStore_

#include "db-inc.h"
#include <qpid/broker/MessageStore.h>
#include <qpid/sys/Monitor.h>
#include <map>
#include <set>
#include <iostream>


namespace qpid {
    namespace bdbstore {
        using std::string;

        class StoreException : public std::exception 
        {
            string text;
        public:
            StoreException(const string& _text) : text(_text) {}
            StoreException(const string& _text, DbException& cause) : text(_text + ": " + cause.what()) {}
            ~StoreException() throw() {}
            const char* what() const throw() { return text.c_str(); }
        };

        class TxnCtxt : public qpid::broker::TransactionContext
        {
            DbTxn* txn;
        public:
            TxnCtxt() : txn(0) {}
            void begin(DbEnv& env){ env.txn_begin(0, &txn, 0); }
            void commit(){ txn->commit(0); txn = 0; }
            void abort(){ txn->abort(); txn = 0; }
            DbTxn* get(){ return txn; }
            virtual bool isTPC() { return false; }
        };

        class TPCTxnCtxt : public TxnCtxt, public qpid::broker::TPCTransactionContext
        {
            const string xid;
        public:
            TPCTxnCtxt(const string& _xid) : xid(_xid) {}
            virtual bool isTPC() { return true; }
            const string& getXid() { return xid; }
        };
            
        class CursorWrapper{
            Dbc* cursor;
        public:
            CursorWrapper() : cursor(0) {}
            ~CursorWrapper() { if(cursor) cursor->close(); }
            void open(Db& db, DbTxn* txn, u_int32_t flags = 0) { db.cursor(txn, &cursor, flags); }
            void close() { if(cursor) cursor->close(); cursor = 0; }
            Dbc* get() { return cursor; }
            Dbc* operator->() { return cursor; }
        };

        class BufferValue : public Dbt
        {
        public:
            qpid::framing::Buffer buffer;

            BufferValue(u_int32_t size, u_int64_t offset);
        };

        class DupKeyRecorder
        {
        protected:
            Db& db;
            const std::string error;
        public:
            DupKeyRecorder(Db& db, const std::string& error);
            void put(DbTxn* txn, Dbt& key, Dbt& value);
        };

        class XidOpRecorder : public DupKeyRecorder
        {
            const std::string name;
        public:
            XidOpRecorder(Db& db, const std::string& name, const std::string& error);
            void put(TPCTxnCtxt& txn, u_int64_t messageId, u_int64_t queueId);
        };

        /**
         * An implementation of the MessageStore interface based on Berkeley DB
         */
        class BdbMessageStore : public qpid::broker::MessageStore
        {
            DbEnv env;
            Db queueDb;
            Db messageDb;
            Db mappingDb;
            Db enqueueXidDb;
            Db dequeueXidDb;
            Db prepareXidDb;
            qpid::sys::Mutex sequenceLock;
            u_int64_t nextQueueId;
            u_int64_t nextMessageId;

            DupKeyRecorder enqueueHelper;
            XidOpRecorder enqueueXidHelper;
            XidOpRecorder dequeueXidHelper;

            typedef std::map<u_int64_t, qpid::broker::RecoverableQueue::shared_ptr> queue_index;

            void recoverQueues(qpid::broker::RecoveryManager& recovery, queue_index& index);
            void recoverMessages(qpid::broker::RecoveryManager& recovery, queue_index& index);
            int enqueueMessage(Dbt& msgId, qpid::broker::RecoverableMessage::shared_ptr& msg, queue_index& index);
            void recoverXids();
            void readXids(Db& db, std::set<string>& xids);
            TxnCtxt* check(qpid::broker::TransactionContext* ctxt);
            u_int64_t getNextQueueId();
            u_int64_t getNextMessageId();
            void store(DbTxn* txn, Dbt& messageId, qpid::broker::PersistableMessage& message);
            void enqueue(DbTxn* txn, Dbt& messageId, Dbt& queueId);
            bool dequeue(DbTxn* txn, Dbt& messageId, Dbt& queueId);
            u_int64_t getRecordSize(Db& db, Dbt& key);
            bool deleteIfUnused(CursorWrapper& cursor, DbTxn* txn, Dbt& messageId);
            bool deleteIfUnused(DbTxn* txn, Dbt& messageId);

            void putNoDupData(DbTxn* txn, Db& db, Dbt& key, Dbt& value, const std::string& error);

            void completed(TPCTxnCtxt& txn, Db& discard, Db& apply);

        public:
            BdbMessageStore(const char* envpath = 0);
            ~BdbMessageStore();

            void truncate();

            void create(const qpid::broker::PersistableQueue& queue);
            void destroy(const qpid::broker::PersistableQueue& queue);

            void create(const qpid::broker::PersistableExchange& queue);
            void destroy(const qpid::broker::PersistableExchange& queue);

            void recover(qpid::broker::RecoveryManager& queues);

            void stage(qpid::broker::PersistableMessage& msg);
            void destroy(qpid::broker::PersistableMessage& msg);
            void appendContent(qpid::broker::PersistableMessage& msg, const std::string& data);
            void loadContent(qpid::broker::PersistableMessage& msg, std::string& data, u_int64_t offset, u_int32_t length);

            void enqueue(qpid::broker::TransactionContext* ctxt, 
                         qpid::broker::PersistableMessage& msg, 
                         const qpid::broker::PersistableQueue& queue);
            void dequeue(qpid::broker::TransactionContext* ctxt, 
                         qpid::broker::PersistableMessage& msg, 
                         const qpid::broker::PersistableQueue& queue);

            void collectPreparedXids(std::set<string>& xids);

            std::auto_ptr<qpid::broker::TransactionContext> begin();
            std::auto_ptr<qpid::broker::TPCTransactionContext> begin(const std::string& xid);
            void prepare(qpid::broker::TPCTransactionContext& ctxt);
            void commit(qpid::broker::TransactionContext& ctxt);
            void abort(qpid::broker::TransactionContext& ctxt);
        };
        string appendLocationInfo(const string& text, const string& file = "", int line = 0);
        string appendBdbError(const string& text, int code);
    }

    extern "C" qpid::broker::MessageStore* create();
    extern "C" void destroy(qpid::broker::MessageStore* store);
}

#define THROW_STORE_EXCEPTION(MESSAGE) throw StoreException(appendLocationInfo(MESSAGE, __FILE__, __LINE__))
#define THROW_STORE_EXCEPTION_2(MESSAGE, EXCEPTION) throw StoreException(appendLocationInfo(MESSAGE, __FILE__, __LINE__), EXCEPTION)


#endif
